@@ -1,6 +1,7 @@
 package main
 
 import (
+	"context"
 	"fmt"
 	"io"
 	"os"
@@ -9,7 +10,7 @@ import (
 	"github.com/la5nta/pat/internal/buildinfo"
 )
 
-func envHandle(_ []string) {
+func envHandle(_ context.Context, _ []string) {
 	writeEnvAll(os.Stdout)
 }
 
@@ -25,8 +26,8 @@ func writeEnvAll(w io.Writer) {
 	writeEnv(w, "PAT_EVENTLOG_PATH", fOptions.EventLogPath)
 	writeEnv(w, "PAT_FORMS_PATH", fOptions.FormsPath)
 	writeEnv(w, "PAT_DEBUG", os.Getenv("PAT_DEBUG"))
+	writeEnv(w, "PAT_WEB_DEV_ADDR", os.Getenv("PAT_WEB_DEV_ADDR"))
 	writeEnv(w, "ARDOP_DEBUG", os.Getenv("ARDOP_DEBUG"))
-	writeEnv(w, "WINMOR_DEBUG", os.Getenv("WINMOR_DEBUG"))
 	writeEnv(w, "PACTOR_DEBUG", os.Getenv("PACTOR_DEBUG"))
 }
 
