@@ -7,10 +7,12 @@ package main
 import (
 	"bufio"
 	"bytes"
+	"context"
 	"embed"
 	"encoding/json"
 	"errors"
 	"fmt"
+
 	"html/template"
 	"io"
 	"io/fs"
@@ -18,6 +20,7 @@ import (
 	"mime/multipart"
 	"net"
 	"net/http"
+	"net/http/httputil"
 	"net/url"
 	"os"
 	"path"
@@ -27,6 +30,8 @@ import (
 	"strings"
 	"time"
 
+	"github.com/la5nta/wl2k-go/transport/ardop"
+
 	"github.com/la5nta/pat/internal/buildinfo"
 	"github.com/la5nta/pat/internal/gpsd"
 
@@ -36,9 +41,10 @@ import (
 	"github.com/la5nta/wl2k-go/fbb"
 	"github.com/la5nta/wl2k-go/mailbox"
 	"github.com/microcosm-cc/bluemonday"
+	"github.com/n8jja/Pat-Vara/vara"
 )
 
-//go:embed web/res/**
+//go:embed web/dist/**
 var embeddedFS embed.FS
 
 var staticContent fs.FS
@@ -84,7 +90,9 @@ func init() {
 	}
 }
 
-func ListenAndServe(addr string) error {
+func devServerAddr() string { return strings.TrimSuffix(os.Getenv("PAT_WEB_DEV_ADDR"), "/") }
+
+func ListenAndServe(ctx context.Context, addr string) error {
 	log.Printf("Starting HTTP service (http://%s)...", addr)
 
 	if host, _, _ := net.SplitHostPort(addr); host == "" && config.GPSd.EnableHTTP {
@@ -94,6 +102,7 @@ func ListenAndServe(addr string) error {
 	}
 
 	r := mux.NewRouter()
+	r.HandleFunc("/api/bandwidths", bandwidthsHandler).Methods("GET")
 	r.HandleFunc("/api/connect_aliases", connectAliasesHandler).Methods("GET")
 	r.HandleFunc("/api/connect", ConnectHandler)
 	r.HandleFunc("/api/formcatalog", formsMgr.GetFormsCatalogHandler).Methods("GET")
@@ -114,15 +123,43 @@ func ListenAndServe(addr string) error {
 	r.HandleFunc("/api/qsy", qsyHandler).Methods("POST")
 	r.HandleFunc("/api/rmslist", rmslistHandler).Methods("GET")
 	r.HandleFunc("/ws", wsHandler)
-	r.HandleFunc("/ui", uiHandler).Methods("GET")
+	r.HandleFunc("/ui", uiHandler()).Methods("GET")
 	r.HandleFunc("/", rootHandler).Methods("GET")
 
 	http.Handle("/", r)
-	http.Handle("/res/", http.FileServer(http.FS(staticContent)))
+	http.Handle("/dist/", distHandler())
 
 	websocketHub = NewWSHub()
 
-	return http.ListenAndServe(addr, nil)
+	srv := http.Server{Addr: addr}
+	errs := make(chan error, 1)
+	go func() {
+		errs <- srv.ListenAndServe()
+	}()
+
+	select {
+	case <-ctx.Done():
+		log.Println("Shutting down HTTP server...")
+		ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
+		defer cancel()
+		srv.Shutdown(ctx)
+		return nil
+	case err := <-errs:
+		return err
+	}
+}
+
+func distHandler() http.Handler {
+	switch target := devServerAddr(); {
+	case target != "":
+		targetURL, err := url.Parse(target)
+		if err != nil {
+			log.Fatalf("invalid proxy target URL: %v", err)
+		}
+		return httputil.NewSingleHostReverseProxy(targetURL)
+	default:
+		return http.FileServer(http.FS(staticContent))
+	}
 }
 
 func rootHandler(w http.ResponseWriter, r *http.Request) {
@@ -304,7 +341,7 @@ func attachFile(f *multipart.FileHeader, msg *fbb.Message) error {
 	// For some unknown reason, we receive this empty unnamed file when no
 	// attachment is provided. Prior to Go 1.10, this was filtered by
 	// multipart.Reader.
-	if isEmptyFormFile(f) {
+	if f.Size == 0 && f.Filename == "" {
 		return nil
 	}
 
@@ -355,23 +392,36 @@ func wsHandler(w http.ResponseWriter, r *http.Request) {
 	websocketHub.Handle(conn)
 }
 
-func uiHandler(w http.ResponseWriter, _ *http.Request) {
-	data, err := fs.ReadFile(staticContent, path.Join("res", "tmpl", "index.html"))
-	if err != nil {
-		log.Fatal(err)
+func uiHandler() http.HandlerFunc {
+	const indexPath = "dist/index.html"
+	templateFunc := func() ([]byte, error) { return fs.ReadFile(staticContent, indexPath) }
+	if target := devServerAddr(); target != "" {
+		templateFunc = func() ([]byte, error) {
+			resp, err := http.Get(target + "/" + indexPath)
+			if err != nil {
+				return nil, fmt.Errorf("dev server not reachable: %w", err)
+			}
+			defer resp.Body.Close()
+			return io.ReadAll(resp.Body)
+		}
 	}
 
-	t := template.New("index.html") // create a new template
-	t, err = t.Parse(string(data))
-	if err != nil {
-		log.Fatal(err)
-	}
-
-	tmplData := struct{ AppName, Version, Mycall string }{buildinfo.AppName, buildinfo.VersionString(), fOptions.MyCall}
-
-	err = t.Execute(w, tmplData)
-	if err != nil {
-		log.Fatal(err)
+	return func(w http.ResponseWriter, _ *http.Request) {
+		data, err := templateFunc()
+		if err != nil {
+			http.Error(w, err.Error(), http.StatusInternalServerError)
+			return
+		}
+		t, err := template.New("index.html").Parse(string(data))
+		if err != nil {
+			http.Error(w, err.Error(), http.StatusInternalServerError)
+			return
+		}
+		tmplData := struct{ AppName, Version, Mycall string }{buildinfo.AppName, buildinfo.VersionString(), fOptions.MyCall}
+		if err := t.Execute(w, tmplData); err != nil {
+			http.Error(w, err.Error(), http.StatusInternalServerError)
+			return
+		}
 	}
 }
 
@@ -400,13 +450,51 @@ func statusHandler(w http.ResponseWriter, _ *http.Request) {
 	_ = json.NewEncoder(w).Encode(getStatus())
 }
 
+func bandwidthsHandler(w http.ResponseWriter, req *http.Request) {
+	type BandwidthResponse struct {
+		Mode       string   `json:"mode"`
+		Bandwidths []string `json:"bandwidths"`
+		Default    string   `json:"default,omitempty"`
+	}
+	mode := strings.ToLower(req.FormValue("mode"))
+	resp := BandwidthResponse{Mode: mode, Bandwidths: []string{}}
+	switch mode {
+	case MethodPactor:
+		fallthrough
+	case MethodAX25:
+		fallthrough
+	case MethodTelnet:
+		fallthrough
+	case MethodVaraFM:
+		fallthrough
+	case MethodSerialTNC:
+		// bandwidth not supported, return empty list
+	case MethodArdop:
+		for _, bw := range ardop.Bandwidths() {
+			resp.Bandwidths = append(resp.Bandwidths, bw.String())
+		}
+		if bw := config.Ardop.ARQBandwidth; !bw.IsZero() {
+			resp.Default = bw.String()
+		}
+	case MethodVaraHF:
+		resp.Bandwidths = vara.Bandwidths()
+		if bw := config.VaraHF.Bandwidth; bw != 0 {
+			resp.Default = fmt.Sprintf("%d", bw)
+		}
+	default:
+		http.Error(w, "mode not recognized", http.StatusBadRequest)
+		return
+	}
+	_ = json.NewEncoder(w).Encode(resp)
+}
+
 func rmslistHandler(w http.ResponseWriter, req *http.Request) {
 	forceDownload, _ := strconv.ParseBool(req.FormValue("force-download"))
 	band := req.FormValue("band")
 	mode := strings.ToLower(req.FormValue("mode"))
 	prefix := strings.ToUpper(req.FormValue("prefix"))
 
-	list, err := ReadRMSList(forceDownload, func(r RMS) bool {
+	list, err := ReadRMSList(req.Context(), forceDownload, func(r RMS) bool {
 		switch {
 		case r.URL == nil:
 			return false
