@@ -6,12 +6,14 @@
 package main
 
 import (
+	"context"
 	"fmt"
 	"io"
 	"log"
 	"net"
 	"os"
 	"os/exec"
+	"os/signal"
 	"path/filepath"
 	"runtime"
 	"strconv"
@@ -33,12 +35,13 @@ import (
 )
 
 const (
-	MethodWinmor    = "winmor"
 	MethodArdop     = "ardop"
 	MethodTelnet    = "telnet"
 	MethodAX25      = "ax25"
 	MethodSerialTNC = "serial-tnc"
 	MethodPactor    = "pactor"
+	MethodVaraHF    = "varahf"
+	MethodVaraFM    = "varafm"
 )
 
 var commands = []Command{
@@ -73,16 +76,17 @@ var commands = []Command{
 		LongLived:  true,
 	},
 	{
-		Str: "compose",
-		Desc: "Compose a new message.\n" +
+		Str:  "compose",
+		Desc: "Compose a new message.",
+		Usage: "[options]\n" +
 			"\tIf no options are passed, composes interactively.\n" +
 			"\tIf options are passed, reads message from stdin similar to mail(1).",
-		Usage: "[options]",
 		Options: map[string]string{
-			"--callsign, -r":    "Callsign to send from. Default reads from config",
+			"--from, -r":        "Address to send from. Default is your call from config or --mycall, but can be specified to use tactical addresses.",
 			"--subject, -s":     "Subject",
 			"--attachment , -a": "Attachment path (may be repeated)",
 			"--cc, -c":          "CC Address(es) (may be repeated)",
+			"--p2p-only":        "Send over peer to peer links only (avoid CMS)",
 			"":                  "Recipient address (may be repeated)",
 		},
 		HandleFunc: composeMessage,
@@ -90,8 +94,8 @@ var commands = []Command{
 	{
 		Str:  "read",
 		Desc: "Read messages.",
-		HandleFunc: func(args []string) {
-			readMail()
+		HandleFunc: func(ctx context.Context, args []string) {
+			readMail(ctx)
 		},
 	},
 	{
@@ -137,8 +141,8 @@ var commands = []Command{
 	{
 		Str:  "updateforms",
 		Desc: "Download the latest form templates from winlink.org.",
-		HandleFunc: func(args []string) {
-			if _, err := formsMgr.UpdateFormTemplates(); err != nil {
+		HandleFunc: func(ctx context.Context, args []string) {
+			if _, err := formsMgr.UpdateFormTemplates(ctx); err != nil {
 				log.Printf("%v", err)
 			}
 		},
@@ -151,7 +155,7 @@ var commands = []Command{
 	{
 		Str:  "version",
 		Desc: "Print the application version.",
-		HandleFunc: func(args []string) {
+		HandleFunc: func(_ context.Context, args []string) {
 			fmt.Printf("%s %s\n", buildinfo.AppName, buildinfo.VersionString())
 		},
 	},
@@ -200,10 +204,9 @@ func optionsSet() *pflag.FlagSet {
 	set := pflag.NewFlagSet("options", pflag.ExitOnError)
 
 	set.StringVar(&fOptions.MyCall, "mycall", "", "Your callsign (winlink user).")
-	set.StringVarP(&fOptions.Listen, "listen", "l", "", "Comma-separated list of methods to listen on (e.g. winmor,ardop,telnet,ax25).")
+	set.StringVarP(&fOptions.Listen, "listen", "l", "", "Comma-separated list of methods to listen on (e.g. ardop,telnet,ax25).")
 	set.BoolVarP(&fOptions.SendOnly, "send-only", "s", false, "Download inbound messages later, send only.")
 	set.BoolVarP(&fOptions.RadioOnly, "radio-only", "", false, "Radio Only mode (Winlink Hybrid RMS only).")
-	set.BoolVarP(&fOptions.Robust, "robust", "r", false, "Use robust modes only (useful to improve s/n-ratio at remote winmor station).")
 	set.BoolVar(&fOptions.IgnoreBusy, "ignore-busy", false, "Don't wait for clear channel before connecting to a node.")
 
 	defaultMBox := filepath.Join(directories.DataDir(), "mailbox")
@@ -257,13 +260,33 @@ func main() {
 	debug.Printf("Event log file is\t'%s'", fOptions.EventLogPath)
 	directories.MigrateLegacyDataDir()
 
+	// Graceful shutdown by cancelling background context on interrupt.
+	//
+	// If we have an active connection, cancel that instead.
+	sig := make(chan os.Signal, 1)
+	signal.Notify(sig, os.Interrupt)
+	ctx, cancel := context.WithCancel(context.Background())
+	defer cancel()
+	go func() {
+		dirtyDisconnectNext := false // So we can do a dirty disconnect on the second interrupt
+		for {
+			<-sig
+			if ok := abortActiveConnection(dirtyDisconnectNext); ok {
+				dirtyDisconnectNext = !dirtyDisconnectNext
+			} else {
+				break
+			}
+		}
+		cancel()
+	}()
+
 	// Skip initialization for some commands
 	switch cmd.Str {
 	case "help":
 		helpHandle(args)
 		return
 	case "configure", "version":
-		cmd.HandleFunc(args)
+		cmd.HandleFunc(ctx, args)
 		return
 	}
 
@@ -325,6 +348,7 @@ func main() {
 		AppVersion: buildinfo.VersionStringShort(),
 		UserAgent:  buildinfo.UserAgent(),
 		LineReader: readLine,
+		GPSd:       config.GPSd,
 	})
 
 	// Make sure we clean up on exit, closing any open resources etc.
@@ -357,11 +381,11 @@ func main() {
 	}
 
 	// Start command execution
-	cmd.HandleFunc(args)
+	cmd.HandleFunc(ctx, args)
 }
 
 func maybeUpdateFormsDir(args []string) {
-	// Backward compatability for config file forms_path
+	// Backward compatibility for config file forms_path
 	formsFlagExplicitlyUsed := false
 	for i := 0; i < len(args); i++ {
 		if strings.HasPrefix(args[i], "--forms") {
@@ -374,7 +398,7 @@ func maybeUpdateFormsDir(args []string) {
 	}
 }
 
-func configureHandle(args []string) {
+func configureHandle(ctx context.Context, args []string) {
 	// Ensure config file has been written
 	_, err := ReadConfig(fOptions.ConfigPath)
 	if os.IsNotExist(err) {
@@ -384,14 +408,14 @@ func configureHandle(args []string) {
 		}
 	}
 
-	cmd := exec.Command(EditorName(), fOptions.ConfigPath)
+	cmd := exec.CommandContext(ctx, EditorName(), fOptions.ConfigPath)
 	cmd.Stdin, cmd.Stdout, cmd.Stderr = os.Stdin, os.Stdout, os.Stderr
 	if err := cmd.Run(); err != nil {
 		log.Fatalf("Unable to start editor: %s", err)
 	}
 }
 
-func InteractiveHandle(args []string) {
+func InteractiveHandle(ctx context.Context, args []string) {
 	var http string
 	set := pflag.NewFlagSet("interactive", pflag.ExitOnError)
 	set.StringVar(&http, "http", "", "HTTP listen address")
@@ -399,20 +423,22 @@ func InteractiveHandle(args []string) {
 	set.Parse(args)
 
 	if http == "" {
-		Interactive()
+		Interactive(ctx)
 		return
 	}
 
+	ctx, cancel := context.WithCancel(ctx)
+	defer cancel()
 	go func() {
-		if err := ListenAndServe(http); err != nil {
+		if err := ListenAndServe(ctx, http); err != nil {
 			log.Println(err)
 		}
 	}()
 	time.Sleep(time.Second)
-	Interactive()
+	Interactive(ctx)
 }
 
-func httpHandle(args []string) {
+func httpHandle(ctx context.Context, args []string) {
 	addr := config.HTTPAddr
 	if addr == "" {
 		addr = ":8080" // For backwards compatibility (remove in future)
@@ -429,12 +455,12 @@ func httpHandle(args []string) {
 
 	promptHub.OmitTerminal(true)
 
-	if err := ListenAndServe(addr); err != nil {
-		log.Fatal(err)
+	if err := ListenAndServe(ctx, addr); err != nil {
+		log.Println(err)
 	}
 }
 
-func connectHandle(args []string) {
+func connectHandle(_ context.Context, args []string) {
 	if args[0] == "" {
 		fmt.Println("Missing argument, try 'connect help'.")
 	}
@@ -461,26 +487,34 @@ func helpHandle(args []string) {
 }
 
 func cleanup() {
+	debug.Printf("Starting cleanup")
+	defer debug.Printf("Cleanup done")
+
+	abortActiveConnection(false)
 	listenHub.Close()
-
-	if wmTNC != nil {
-		if err := wmTNC.Close(); err != nil {
-			log.Fatalf("Failure to close winmor TNC: %s", err)
-		}
-	}
-
 	if adTNC != nil {
 		if err := adTNC.Close(); err != nil {
 			log.Fatalf("Failure to close ardop TNC: %s", err)
 		}
 	}
-
 	if pModem != nil {
 		if err := pModem.Close(); err != nil {
 			log.Fatalf("Failure to close pactor modem: %s", err)
 		}
 	}
 
+	if varaFMModem != nil {
+		if err := varaFMModem.Close(); err != nil {
+			log.Fatalf("Failure to close varafm modem: %s", err)
+		}
+	}
+
+	if varaHFModem != nil {
+		if err := varaHFModem.Close(); err != nil {
+			log.Fatalf("Failure to close varahf modem: %s", err)
+		}
+	}
+
 	eventLog.Close()
 }
 
@@ -541,7 +575,7 @@ func loadHamlibRigs() map[string]hamlib.VFO {
 	return rigs
 }
 
-func extractMessageHandle(args []string) {
+func extractMessageHandle(_ context.Context, args []string) {
 	if len(args) == 0 || args[0] == "" {
 		panic("TODO: usage")
 	}
@@ -581,7 +615,7 @@ func EditorName() string {
 	return "vi"
 }
 
-func posReportHandle(args []string) {
+func posReportHandle(ctx context.Context, args []string) {
 	var latlon, comment string
 
 	set := pflag.NewFlagSet("position", pflag.ExitOnError)
@@ -614,15 +648,28 @@ func posReportHandle(args []string) {
 			log.Fatalf("GPSd daemon: %s", err)
 		}
 		defer conn.Close()
-
 		conn.Watch(true)
 
-		log.Println("Waiting for position from GPSd...") // TODO: Spinning bar?
-		pos, err := conn.NextPos()
-		if err != nil {
-			log.Fatalf("GPSd: %s", err)
-		}
+		posChan := make(chan gpsd.Position)
+		go func() {
+			defer close(posChan)
+			pos, err := conn.NextPos()
+			if err != nil {
+				log.Printf("GPSd: %s", err)
+				return
+			}
+			posChan <- pos
+		}()
 
+		log.Println("Waiting for position from GPSd...") // TODO: Spinning bar?
+		var pos gpsd.Position
+		select {
+		case p := <-posChan:
+			pos = p
+		case <-ctx.Done():
+			log.Println("Cancelled")
+			return
+		}
 		report.Lat = &pos.Lat
 		report.Lon = &pos.Lon
 		if config.GPSd.UseServerTime {
